
/* nxlcau22.c				*/
/* (2,2) Linear Cellular Automaton	*/

/* Reference:				*/
/*					*/
/*    Kenneth E. Perry			*/
/*    Abstract Mathematical Art		*/
/*    BYTE				*/
/*    December, 1986			*/
/*    pages 181-192			*/

/*    Copyright (C) 1987		*/
/*    Copyright (C) 1988		*/
/*    Copyright (C) 1989		*/
/*    Copyright (C) 1994		*/
/*    Harold V. McIntosh		*/
/*    Gerardo Cisneros S.		*/

/* G. Cisneros, 4.3.87						*/
/* 28 April 1987 - version for XVI Feria de Puebla [HVM]	*/
/* 14 May 1987 - modified for (3,1) and general rule [HVM]	*/
/* 19 May 1987 - modified for (2,2) [HVM]			*/
/* 20 May 1987 - modified for (2,3) [HVM]			*/
/* 21 May 1987 - Wolfram rule number [HVM]			*/
/* 8 June 1987 - general rule for (4,1) [HVM]		 	*/
/* 12 June 1987 - cartesian product of (2,1) rules [HVM]	*/
/* 12 June 1987 - (2,1) rule with memory  [HVM]			*/
/* 14 June 1987 - individual cycles of evolution [HVM]		*/
/* 17 June 1987 - p-adic representation in plane [HVM]		*/
/* 22 June 1987 - 2 speed gliders via 16 transitions [HVM]	*/
/* 26 June 1987 - push, pop the rule [HVM]			*/
/* 26 June 1987 - conserve position of rule cursor [HVM]	*/
/* 27 June 1987 - incremental rule construction [HVM]		*/
/* 29 June 1987 - conserve position of cell pointer [HVM]	*/
/* 30 June 1987 - mark & unmark transitions, xchg x&X [HVM]	*/
/* 25 July 1987 - display and edit de Bruijn diagrams [HVM]	*/
/* 27 July 1987 - graph of transition probabilities [HVM]	*/
/* 4 September 1987 - PROB21.C for option 't' [HVM]		*/
/* 21 October 1987 - program disks disappeared			*/
/* 20 December 1987 - program reconstructed from listings	*/
/* 20 February 1988 - RIJN21.C for option 'd' [HVM]		*/
/* 11 March 1989 - x option in trule [HVM]			*/
/* 16 March 1989 - *+- option to change screen color [HVM]	*/
/* 17 September 1989 - directory for demonstration rules [HVM]	*/
/* 15 October 1989 - uniformize edrul, edline with lcau31 [HVM]	*/
/* 18 October 1989 - rule search for prescribed evolution [HVM]	*/
/* 10 February 1990 - Garden of Eden option e			*/
/* 21 June 1994 - adaptation to NeXT				*/

/* # include <bdos.h> */
#import "nxlcau22.h" 
#import "rec.h" 
#import "rectbl.h" 
#import "keybrd.h" 


# define LROW		11	/* top row of line panel     */
# define UROW		 5	/* bottom line of rule panel */



/* extern Inst  xprg[][128];   */
extern struct fptbl dtbl2[];
extern char  idef[][30];
extern char recrule[][CLEN2];

Inst xprg[10][XLEN];


char ixrule[NX][DS]= {

	"00010110011010010110100110010111",	/* totalistic rule #52	*/
	"00010000011101001011001011101111",	/* half and half #1   	*/
	"00010110011010010110100110010110",	/* totalistic rule #20	*/
	"00011100010000000010100010000000"	/* Life diagonal ripple */
	"00100100110010110110111000011001",	/* (1,7) glider		*/
	"00100110101000110001010100110110",	/* honeycomb		*/
	"00101010001110110011111101100110",	/* Li figure 1 comp    	*/
	"01010000011101001011001011101110",	/* half and half #2	*/
	"01101000111010100011001011010101",	/* half and half #3	*/
	"10000100011100110111100101110010",	/* Li figure 2          */
	"10010000001111001110100110110001",	/* another fourish	*/
	"10011000010001011001011110110011",	/* Li figure 6		*/
	"10011001000000110010001110101011",	/* Li figure 1		*/
	"10110001010001010011000111011110",	/* Li figure 4         	*/
	"11000100101111100000101010011001",	/* class fourish	*/
	"11010010111100001101001011110000",	/* reversible rule   	*/
	"11110000101101001111000010110100",	/* its reverse        	*/
	"11111111000000001100111100110000"	/* reversible [(2,3)]	*/

	};
char recrule2[NY2][CLEN2]={
          "(\"3918960232\"W; [1 - Totalistic rule 52 ] )",
	  "(\"4149030408\"W; [2 - Half and Half # 1  ] )",
	  "(\"1771476584\"W; [3 - Totalistic rule # 20 ])",
	  "(\"18088504\"W;   [4 - Life diagonal ripple  ] )",
	  "(\"2557924132\"W; [5 - (1,7) glider  ] )",
	  "(\"1822999908\"W; [6 - Honeycomb  ] )",
	  "(\"1727847508\"W; [7 - Li figure 1 comp  ] )",
	  "(\"2001546762\"W; [8 - Half and Half # 2  ] )",
	  "(\"2873906966\"W; [9 - Half and Half # 3  ] )",
	  "(\"1319030305\"W; [10 - Li figure 2 ] )",
	  "(\"2375498761\"W; [11 - Another fourish ] )",
	  "(\"3454640665\"W; [12 - Li figure 6  ] )",
	  "(\"3586441369\"W; [13 - Li figure 1  ] )",
	  "(\"2072814221\"W; [14 - Li figure 4  ] )",
	  "(\"2572188963\"W; [15 - Class fourish ] )",
	  "(\"D2F0D2F0\"W;  [16 - Reversible rule  ] )",
	  "(\"755969295\"W; [17 - Its reverse  ] )",
	  "(\"217252095\"W; [18 - reversible [(2,3)] ] )"

        };
 
parneih amat, bmat, fmat, gmat, hmat, xmat;
twoparn smat, tmat, umat, vmat;


char   ttrule[TS];
char   tarule[TS];
char   strule[KK][TS-1];
char   sarule[KK][TS-1];
char   gf2rule[13];			/*   coefficients of GF2 rule	*/
char   ascrule[KK][KK][KK][KK][KK];
char   auxrule[KK][KK][KK][KK][KK];	/* auxiliary transition table */
int    binrule[KK][KK][KK][KK][KK];
char   prule1[KK][KK][KK];
char   prule2[KK][KK][KK];		/* composite rule specification */
char   arr1[AL], arr2[AL], arr3[AL];
int    ru, ru0, ru1, ru2, ru3, ru4, ru5, ru6, ru7, ru8;	/* marker in edrule        */
int    co=1, dn=1;				/* connectivity, dimension */
int    pr;					/* probability, in mils    */
int    ix0, iy0;				/* origin for pen moves    */
int    dij, di, dj;				/* demo pointer            */
int    rij, rdi, rdj;				/* rec demo pointers 	   */
int    li, lj;					/* cursor for edline       */
int    pr, pl;					/* rule, line prob, mils   */
int    rp, rq;					/* source code in cstr	   */
int    bi;					/* bernstein index	   */
int    ct0, ct1, ct2, ct3;			/* general counters	   */
int    sw0, sw1, sw2, sw3;			/* general switches	   */
int    ex, wy;					/* hiplot coordinates	   */
double nu0, nu1, nu2, nu3;			/* general numeric values  */
double mf;					/* mouse increment	   */
double sm[TS];					/* selfconsistent nbrhd    */
double sn[TS];					/* selfconsistent nbrhd    */
double bx[BD];					/* bernstein factor        */
double bp[BD];
double pv[10];					/* polynomial vector 	   */
double fpa, fpb, fpc;				/* fixed points		   */
double blpr[KK];			/* probability from blocks	*/

Inst   rprg[900];			/* rec program array */
char   cstr[CLEN];			/* console rec program */
char   cprg[CLEN];			/* edit stored program */

/*extern char recrule[NY][48];*/

long int r_intpar;


nxlcau22main()
{
int  i, j, i0, i1, i2, i3, i4;
int  more='r', sc=0;
char a, b, c, p0, p1, p2;

for (i=0; i<TS; i++) ttrule[i]='0';

ru=6; ru0=ru1=ru2=ru3=ru4=0;
li=SL/2; lj=0;					/* line editor cursor */
auxblnk(auxrule);
ct0=ct1=ct2=ct3=0;				/* counters	       */
sw0=sw1=sw2=sw3=0;				/* switches	       */
co=1;						/* connectivity      */
dn=1;						/* dimension         */
rij=2; rdi=0; rdj=1;
rp=rq=0; cstr[rp]=0;
pr=500;	pl=500;					/* probability, mils */
mf=0.05;
for (i=0; i<BD; i++) bx[i]=1.0;			/* bernstein factor */


    videopalette(WHCYMAG);			/* white/cyan/magenta */

    tuto();
    while (!kbdst()) rand();			/* wait for keypress */
    kbdin();					/* ignore it */
    videomode(T80X25);
    videoscroll(0,0,CR,71,0,3);			/* menu on blue background */
    videoscroll(CL+8,0,24,71,0,3);
    i=(rand()%NX);
    di=i%DROWS;
    dj=(i/DROWS==1?DCOLS:1);
    xtoasc(ascrule,i);
    ranlin();					/* random initial array */

    while (more!='n') {				/* execute multiple runs */
    rmenu();
    lmenu(); 
    while (0<1) {				/* set up one run */
    c=kbdin();
    if (c=='g') break;				/* go draw graph */
    if (c=='q') more='n';			/* quit for good */
    if (more=='n') break;
    switch (c) {

        case Kf1:				/* F1 = fetch stored rule */
	    dij=dmenu(Kf1);
	    xtoasc(ascrule,dij);
	    rmenu();
	    lmenu();
	    break;
        case Kf2:				/* F2 = rec program */
	    rij=lim(1,recdem(Kf2),NY)-1;
	    for (i=0; i<CLEN; i++) cstr[i]=recrule[rij][i];
	    rmenu();
	    lmenu();
	    break;
        case Kf3:				/* f3 = edit rec program */
	    edrec(cstr); 
	    rmenu();
	    lmenu();
	    break;
        case Kf4:				/* f4 = execute rec program */
	    if (rec_c(cstr,rprg,PLEN,dtbl)==1) rec_x(rprg);
	    break;
        case Kf5:				/* f5 = execute demo program */
	    if (rec_c(&recrule[rdi],rprg,PLEN,dtbl)==1) rec_x(rprg);
	    break;

	case '*': sc=2; break;			/* white on red */
	case '+': sc=1; break;			/* white screen */
	case '-': sc=0; break;			/* blue screen */
	case '@':				/* numbered tot rule */
	    nutoto(numin(0));
	    totoasc();
	    rmenu();
	    videocursor(0,CM,0);
	    break;
	case '$':				/* dozen totalistics */
	    j=numin(0);
	    for (i=0; i<12; i++) {
	      nutoto(j+i);
	      totoasc();
	      ranlin();
	      evolve(sc);
	      }
            videomode(T80X25);
	    rmenu();
	    lmenu();
	    break;
        case 'o': co=1; break;		/* random neighbors    */
        case 'O': co=0; break;		/* normal connectivity */
	case 'p':			/* product of two (2,1) rules */
	    nutowo(numin(0),prule1);
	    nutowo(numin(0),prule2);
	    for (i0=0; i0<KK; i0++)
	    for (i1=0; i1<KK; i1++)
	    for (i2=0; i2<KK; i2++)
	    for (i3=0; i3<KK; i3++)
	    for (i4=0; i4<KK; i4++)
	      {
	      p0=prule1[i0][i1][i2]-'0';
	      p1=prule1[i1][i2][i3]-'0';
	      p2=prule1[i2][i3][i4]-'0';
	      ascrule[i0][i1][i2][i3][i4]=prule2[p0][p1][p2];
	      }
	    xblnk();
	    rmenu();
	    break;
	case 'T':				/* totalistic rule */
	    xblnk();
	    tmenu();
	    edtrule();
	    totoasc();
	    for (i0=0; i0<KK; i0++)
	    for (i1=0; i1<KK; i1++)
	    for (i2=0; i2<KK; i2++)
	    for (i3=0; i3<KK; i3++)
	    for (i4=0; i4<KK; i4++)
	      ascrule[i0][i1][i2][i3][i4]=ttrule[i0+i1+i2+i3+i4];
	    videocursor(0,CM,0);
	    rmenu();
	    break;
	case 't':					/* probability calculation */
	    edtri();
	    rmenu();
	    lmenu();
	    break;
        case 'r':					/* edit rule */	
	    xblnk();
	    edrule();
	    videocursor(0,CM,0);
	    rmenu();
	    break;
        case 'l':					/* edit cell string */
	    xblnk();
	    edline(8,40);
	    videocursor(0,CM,0);
	    lmenu();
	    break;
        case '#':				/* read stored rule */
	    xmenu(NX);
	    xtoasc(ascrule,lim(1,dmenu(),NX)-1);
	    rmenu();
            break;
	case 'D':				/* run through samples */
	    for (i=0; i<NX; i++) {
	      xmenu(i);
	      xtoasc(ascrule,i);
	      ranlin();
	      evolve(sc);
	      }
	    videomode(T80X25);
	    rmenu();
	    break;
	case 'd':				/* de Bruijn diagram */
	    edijn();
	    rmenu();
	    lmenu();
	    break;
	case 'e':				/* Garden of Eden */
	    edard();
	    rmenu();
	    lmenu();
	    break;
        case 'i':
            for (i=0; i<8; i++) {
	    ascincr(ascrule,auxrule);
	    ranlin();
	    evolve(sc);}
	    break;
        case 'u':					/* sparse init arry */
	    xblnk();
	    for (i=0; i<AL; i++) arr1[i]=0;
	    arr1[AL/4]=1;
            arr1[AL/2]=1;
            arr1[(3*AL)/4]=1;
            arr1[(3*AL)/4+2]=1;
	    lmenu();
            break;
	case 'w':					/* Wolfram rule # */
	    i=numin(0);
	    wmenu(i);
	    for (i0=0; i0<KK; i0++) {
	    for (i1=0; i1<KK; i1++) {
	    for (i2=0; i2<KK; i2++) {
	    for (i3=0; i3<KK; i3++) {
	    for (i4=0; i4<KK; i4++) {
	      ascrule[i0][i1][i2][i3][i4]='0'+i%KK;
	      i/=KK;
	    }}}}}
	    rmenu();
	    break;
	case 'x':					/* random rule */
	    xblnk();
	    i=rand();
	    for (i0=0; i0<KK; i0++) {
	    for (i1=0; i1<KK; i1++) {
	    for (i2=0; i2<KK; i2++) {
	    for (i3=0; i3<KK; i3++) {
	    for (i4=0; i4<KK; i4++) {
	      if (i==0) i=rand();
	      if (auxrule[i0][i1][i2][i3][i4]==' ') {
		ascrule[i0][i1][i2][i3][i4]='0'+i%KK;
		i/=KK;
		}
	      }}}}}
	    rmenu();
	    break;
	case 'X':					/* random rule */
	    xblnk();
	    i=rand();
	    for (i0=0; i0<KK; i0++) {
	    for (i1=0; i1<KK; i1++) {
	    for (i2=0; i2<KK; i2++) {
	    for (i3=0; i3<KK; i3++) {
	    for (i4=0; i4<KK; i4++) {
	      if (i == 0) i=rand();
	      ascrule[i0][i1][i2][i3][i4]='0'+i%KK;
	      i/=KK;
	    }}}}}
	    rmenu();
	    break;
	case 'y':					/* random line */
	    xblnk();
	    ranlin();
            lmenu();
	    break;
	case 'Y':					/* symmetrize rule */
	    for (i0=0; i0<KK; i0++) {
	    for (i1=0; i1<KK; i1++) {
	    for (i2=0; i2<KK; i2++) {
	    for (i3=0; i3<KK; i3++) {
	    for (i4=0; i4<KK; i4++) {
	    ascrule[i4][i3][i2][i1][i0]=ascrule[i0][i1][i2][i3][i4];      
	    }}}}}
	    break;
	case 'B':					/* begin barrier */
	    a=kbdin();
	    b=kbdin();
	    ascrule[0][a-'0'][b-'0'][a-'0'][a-'0']=a;
	    ascrule[1][a-'0'][b-'0'][a-'0'][a-'0']=a;
	    ascrule[2][a-'0'][b-'0'][a-'0'][a-'0']=a;
	    rmenu();
	    break;
	case 'E':					/* end barrier */
	    a=kbdin();
	    b=kbdin();
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][0]=b;
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][1]=b;
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][2]=b;
	    rmenu();
	    break;
	case 'L':					/* left glider link */
	    a=kbdin();
	    b=kbdin();
	    c=kbdin();
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][c-'0']=c;
	    rmenu();
	    break;
	case 'R':					/* left glider link */
	    a=kbdin();
	    b=kbdin();
	    c=kbdin();
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][c-'0']=a;
	    rmenu();
	    break;
	case 's':					/* still life link */
	    a=kbdin();
	    b=kbdin();
	    c=kbdin();
	    ascrule[a-'0'][a-'0'][a-'0'][b-'0'][c-'0']=b;
	    rmenu();
	    break;
	case 'S':					/* semitotalistic rule */
	    xblnk();
	    smenu();
	    edsrule();
	    sttoasc();
	    videocursor(0,CM,0);
	    rmenu();
	    break;
	case '.':					/* one cycle of evolution */
	    asctobin();
	    onegen(AL);
	    lmenu();
	    break;
	case '=':
	    for (i=1; i<8;  i++) {
	    for (j=0; j<40; j++) arr1[40*i+j]=arr1[j];}
	    lmenu();
	    break;
	case '~':
	    for (i=1; i<16;  i++) {
	    for (j=0; j<20; j++) arr1[20*i+j]=arr1[j];}
	    lmenu();
	    break;
        default: break;
        }
    }
    if (more=='n') break;
    do {
    evolve(sc);
    videocursor(0,0,0); scrstr("?");
    videocursor(0,0,34); scrstr("y/n/cr");
    more=kbdin();
    } while (more=='\015');
    videomode(T80X25);					/* reset the screen */
    if (more=='n') break;
    }
  videomode(T80X25);}    

/* edit the rule */
edrule() {char c;
  while (0<1) {
    cscrrul();
    videocursor(0,UROW,ru);
    c=kbdin();
    if (c=='\015') break;				/* carriage return exits */
    switch (c) {
    case '0': case '1':					/* state */
      ascrule[ru0][ru1][ru2][ru3][ru4]=c;
      videocattr(0,c,3,1);
      rupl();
      break;
    case '\011':					/* tab = next quad */
      ru4=0; ru1=0; ru0++;
      if (ru0==KK) ru0=0;
      ru=KK*KK*KK*KK*ru0+6;
      break;
    case ' ': case '\315': rupl(); break;				 /* advance */
    case '\010': case '\313': rumi(); break;				 /* backspace */
    case '\013': case '\310': auxrule[ru0][ru1][ru2][ru3][ru4]='+'; rupl(); break; /* up arrow */
    case '\012': case '\320': auxrule[ru0][ru1][ru2][ru3][ru4]=' '; rupl(); break; /* down arrow */
    case 'z': asczero(ascrule);
    default: break;
    }
  }
}

/* advance the rule cursor */
rupl() {
  ru4++;
  if (ru4==KK) {ru4=0; ru3++;}
  if (ru3==KK) {ru3=0; ru2++;}
  if (ru2==KK) {ru2=0; ru1++;}
  if (ru1==KK) {ru1=0; ru0++;}
  if (ru0==KK) {ru0=0; }
  ru=KK*(KK*(KK*(KK*ru0+ru1)+ru2)+ru3)+ru4+6;
  }

/* retract the rule cursor */
rumi() {
  if (ru4!=0) ru4--; else {ru4=KK-1;
  if (ru3!=0) ru3--; else {ru3=KK-1;
  if (ru2!=0) ru2--; else {ru2=KK-1;
  if (ru1!=0) ru1--; else {ru1=KK-1;
  if (ru0!=0) ru0--; else {ru0=KK-1;
    }}}}}
  ru=KK*(KK*(KK*(KK*ru0+ru1)+ru2)+ru3)+ru4+6;
  }

/* edit totalistic rule */
edtrule() {char c; int  i, k;
  i=0;
  while (i<TS) {
    c=ttrule[i];
    videocursor(0,3,56+i);
    videocattr(0,c,3,1);
    c=kbdin();
    if (c=='\015') break;
    switch (c) {
      case '0': case '1':				/* state */
        ttrule[i]=c;
        videocattr(0,c,CR,1);
        i++;
        break;
      case ' ': case '\315':			/* space = advance */
        i++;
        break;
      case '\010': case '\313':			/* backspace */
        if (i!=0) i--;
	break;
      case 'x':					/* random selection */
	for (i=0; i<TS; i++) {
	  if (i%8==0) k=rand();
	  ttrule[i]='0'+k%KK; k/=KK;}
	i=0;
	tscrrul();
	break;
      default: break;
      }
  }
}

/* edit semitotalistic rule */
edsrule() {char c; int  i, j, k;
for (j=0; j<KK; j++) {
  i=0;
  while (i<TS-1) {
    c=strule[j][i];
    videocursor(0,CR-2+j,57+i);
    videocattr(0,c,3,1);
    c=kbdin();
    if (c=='\015') break;
    switch (c) {
      case '0': case '1':			/* state */
        strule[j][i]=c;
        videocattr(0,c,CR,1);
        i++;
        break;
      case ' ': case '\014': case '\315':	/* space = advance */
        i++;
        break;
      case '\010': case '\313':			/* backspace */
        if (i!=0) i--;
	break;
      case '\013': case '\310': if (j==0) j=KK; j--; break; /* up arrow */
      case '\012': case '\320': j++; if (j==KK) j=0; break; /* down arrow */
      case 'x':					/* random selection */
	for (i=0; i<TS-1; i++) {
	  if (i%8==0) k=rand();
	  strule[j][i]='0'+k%KK; k/=KK;}
	i=0;
	sscrrul(j);
	break;
      default: break;
      } /* end switch */
  }     /* end while  */
}       /* end for    */
}

/* edit the line of cells */
edline(nr,nc) int nr, nc; {
char c;
int  i, j, jj, k, l, m, s, sa, sb, t;
int  xx[RR+1], pd[KK*KK*KK*KK*KK+1][2];
int  ll;					/* line to be refreshed */

  videocursor(0,LROW+9,0);
  scrstr("insert states with 0, 1; move cursor with");
  videocursor(0,LROW+10,0);
  scrstr("keyboard arrows, space or backspace.");
  videocursor(0,LROW+11,0);
  scrstr("carriage return exits");
  while (0<1) {
  cscrrul();
  s=0; t=0;
  m=0; /* refresh trigger */
  ll=lj;
  videocursor(0,lj+LROW,li);
  c=kbdin();
  if (c == '\015') {videoscroll(LROW+9,0,LROW+12,70,0,3); break;}
  switch (c) {
  case '0': case '1':					/* enter state */
    arr1[nc*lj+li]=c-'0';
    li++; if (li>=nc) {li=0; lj++;} if (lj>=nr) lj=0;
    break;
  case '\012': case '\320':				/* downwards */
    lj++; if (lj>=nr) lj=0;
    break;
  case '\013': case '\310':				/* upwards */
    if (lj>0) lj--; else lj=nr-1;
    break;
  case '\014': case '\315': case ' ':			/* forward */
    li++; if (li>=nc) {li=0; lj++;} if (lj>=nr) lj=0;
    break;
  case '\010': case '\313':				/* backward */
    if (li>0) li--;
      else {li=nc-1; if (lj>0) lj--; else lj=nr-1;}
    break;
  case '<':						/* left margin */
    li=0;
    break;
  case '>':						/* right margin */
    li=nc-1;
    break;
  case 'z':						/* clear single row */
    for (k=0; k<nc; k++) arr1[nc*lj+k]=0;
    break;
  case 'Z':						/* clear full array */
    for (k=0; k<AL; k++) arr1[k]=0;
    lmenu();
    break;
  case '^':						/* copy line above */
    if (lj>0) for (k=nc*lj; k<nc*(lj+1); k++) arr1[k]=arr1[k-nc];
    lj++; if (lj>=nr) lj=0;
    break;
  case 'x':						/* unmark transitions */
    auxblnk(auxrule);
    cscrrul();
    break;				
  case 'q':						/* clean up the area */
    ulmenu(3);
    break;
  case '=':						/* insert transition */
    if (lj==0) break;
    linix(xx,li,lj,nc);
    c=xx[RR]+'0';
    ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=c;
    videocursor(0,lj+LROW,li); videocattr(0,c=='0'?'.':c,3,1);
    m=1;
    break;
  case '+':		/* set marker for transition in auxrule */
    if (lj==0) break;
    linix(xx,li,lj,nc);
    c=xx[RR]+'0';
    auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
    videocursor(0,lj+LROW,li); videocattr(0,c=='0'?'.':c,3,1);
    m=1;
    break;
  case '-':		/* remove marker for transition from auxrule */
    if (lj==0) break;
    linix(xx,li,lj,nc);
    c=xx[RR]+'0';
    auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=' ';
    videocursor(0,lj+LROW,li); videocattr(0,c=='0'?'.':c,3,1);
    break;
  case '*':		/* insert transitions for the whole line */
    if (lj==0) break;
    for (i=0; i<nc; i++) {
      linix(xx,i,lj,nc);
      c=xx[RR]+'0';
      ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=c;
      auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
      videocattr(0,c=='0'?'.':c,3,1);
      }
    videocursor(0,lj+LROW,li);
    m=1;
    break;
  case '.':		/* point evolution */
    if (lj==0) break;
    linix(xx,li,lj,nc);
    arr1[li+nc*lj]=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
    break;
  case '?':		/* evolution of the whole line */
    if (lj==nr-1) {for (i=0; i<AL-nc; i++) arr1[i]=arr1[i+nc]; lj--; lmenu(); }
    lj++; ll=lj;
    for (i=0; i<nc; i++) {
      linix(xx,i,lj,nc);
      arr1[i+nc*lj]=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
      }
    break;
  case '/':		/* conditional evolution */
    if (lj==nr-1) {for (k=0; k<AL-nc; k++) arr1[k]=arr1[k+nc]; lj--; lmenu(); }
    lj++; ll=lj;
    for (i=0; i<nc; i++) {
      j=0;
      linix(xx,i,lj,nc);
      for (k=0; k<RR; k++) if (xx[k]>KK-1) j=KK;
      if (auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==' ') j=KK;
      if (j==0) j=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
      arr1[i+nc*lj]=j;
      }
    break;
  case 'c':		/* test consistency everywhere */
    for (i=0; i<nc; i++) {
      videocursor(0,LROW,i);
      k=arr1[i];
      videocattr(0,k==0?'.':k+'0',7,1);
      }
    for (j=1; j<nr; j++) {
    for (i=0; i<nc; i++) {
      linix(xx,i,j,nc);
      k=xx[RR]+'0';
      videocursor(0,j+LROW,i);
      videocattr(0,k=='0'?'.':k,k==ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]?7:4,1);
      }}
    m=1;
    break;
  case 'C':		/* test consistency for marked neighborhoods */
    for (i=0; i<nc; i++) {
      videocursor(0,LROW,i);
      k=arr1[i];
      videocattr(0,k==0?'.':k+'0',7,1);
      }
    for (j=1; j<nr;  j++) {
    for (i=0; i<nc; i++) {
      linix(xx,i,j,nc);
      k=xx[RR]+'0';
      videocursor(0,j+LROW,i);
      videocattr(0,k=='0'?'.':k,
        (k==ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]] && 
		auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]!=' ')?7:4,1);
      }}
    m=1;
    break;
  case 'm':		/* try to match a given evolution */
    videocursor(0,23,10); printf("           ");
    s=0; sa=0; sb=0; t=0;
    if (lj==0) break;
/* in each generation -- that is, on each line -- set all unmarked	*/
/* transitions to zero and push the location of the corresponding cell	*/
    if (lj>1) {
      for (j=1; j<lj; j++) {
      for (i=0; i<nc; i++) {
        linix(xx,i,j,nc);
        if (auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==' ') {
          ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='0';
          auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
          s++; pd[s][0]=i; pd[s][1]=j;}
        arr1[i+nc*j]=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
        }}
      }
    cscrrul();
/* Until the target line evolves from the tentative evolution,	*/
/* increment the last recorded transition and repeat the trial.	*/
/* But first, test the target line.				*/
    do {t=0;
    sa=s;
    for (i=k=0; i<nc; i++) {
      linix(xx,i,lj,nc);
      c=xx[RR]+'0';
      if (ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==c &&
	  auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]!=' ') continue;
      if (auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==' ') {
          ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=c;
          auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
          s++; pd[s][0]=i; pd[s][1]=lj; continue;}
      k=1;
      }
    sb=s;
    if (k==0) {t=1; break;}
/* undo last minute definitions, which are otherwise kept.	*/
    while (pd[s][1]==lj) {
      linix(xx,pd[s][0],pd[s][1],nc);
      auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=' ';
      ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='0';
      s--;}
/* Given a mismatch, start backtracking; as far as necessary to	*/
/* a cell which has not run through its cycle.			*/
    if (s==0) {t=2; break;}
    do {
      i=pd[s][0]; j=pd[s][1];
      linix(xx,i,j,nc);
      c=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]];
      c+=1; k=c-'0';
      ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=c;
      auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';	/*---*/
      arr1[i+nc*j]=k;
      if (k<KK) break;
      auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]=' ';
      ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='0';
      arr1[i+nc*j]=0;
      s--;
      } while (s>0);
    if (s==0) {t=2; break;}
/* Once again, run through the trial evolution. First, finish	*/
/* out the line where backtracking stopped.			*/
    for (i=pd[s][0],jj=pd[s][1]; i<nc; i++) {
      linix(xx,i,jj,nc);
      if (auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==' ') {
        ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='0';
        auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
        s++; pd[s][0]=i; pd[s][1]=jj;}
      arr1[i+nc*jj]=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
      }
/* Then redo the remaining lines, if any.			*/
    jj++; if (jj<lj) {
    for (j=jj; j<lj; j++) {
    for (i=0;  i<nc; i++) {
      linix(xx,i,j,nc);
      if (auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]==' ') {
        ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='0';
        auxrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]='+';
        s++; pd[s][0]=i; pd[s][1]=j;}
      arr1[i+nc*j]=ascrule[xx[0]][xx[1]][xx[2]][xx[3]][xx[4]]-'0';
      }} }
/* display progress and test for a waiting interrupt request	*/
    cscrrul();
    lmenu();

    videocursor(0,20,50); printf("(%2d,%2d) %2d %2d",pd[s][1],pd[s][0],sa,sb);
/*    c=kbdin(); if (c=='x') break; */

    if (kbdst()) {kbdin(); t=3; break;}
    } while (t==0);
    videocursor(0,23,10);
    if (t==1) printf("rule found ");
    if (t==2) printf("impossible ");
    if (t==3) printf("interrupted");
    m=1;
    break;
  default: break;
      } /* end switch */
  if (m==0) for (k=0; k<nc; k++) {
    videocursor(0,ll+LROW,k);
    l=arr1[k+nc*ll]+'0';
    videoputc(l=='0'?'.':l,1);
    }
  }     /* end while */
}       /* end edline */

/* calculate index, or rather the neighbors, in edline. */
linix(xx,i,j,nc) int xx[RR+1], i, j, nc; {int k, x;
  xx[(RR-1)/2]=x=i+nc*(j-1);
  if ((i==0)||(i==1)) xx[0]=x+nc-2; else xx[0]=x-2;
  if (i==0) xx[1]=x+nc-1; else xx[1]=x-1;
  if (i==nc-1) xx[3]=x-nc+1; else xx[3]=x+1;
  if ((i==nc-1)||(i==nc-2)) xx[4]=x-nc+2; else xx[4]=x+2;
  for (k=0; k<RR; k++) xx[k]=arr1[xx[k]];
  xx[RR]=arr1[x+nc];
}

/* display a screen of evolution */
void evolve(sc) int sc; {
int i, j;
 // savearry();
  asctobin();
/*  for (j=0; j<200; j++) videodot(j,AL-1,2); */
  for (j=199; j>=0; j--) {

/* evolve for 192 generations 
    switch (sc) {
     case 0: 
  serms((double)(2*i),(double)(2*j),arr1[i],3); */


for (i=0; i<AL; i++)
            videodot(j,i,arr1[i]);

        /*    break;
		case 1: for (i=0; i<AL; i++)videodot(j,i,arr1[i]*3);break;*/
		/*case 2: for (i=0; i<AL; i++)videodot(j,i,arr1[i]+2);break;
      default: break; }  */

    onegen(AL);
    }
//  renewarry();
}

/* copy ascrule over to binrul */ 
asctobin() {int i0, i1, i2, i3, i4;
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
      binrule[i0][i1][i2][i3][i4]=ascrule[i0][i1][i2][i3][i4]-'0';
}

/* evolution for one generation */


onegen(j) int j; {int i, k, i0, i1, i2, i3, i4; long int ii;
switch (dn) {

case 1:
if (j>4) {
  arr2[0]=binrule[arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]][arr1[2]];
  arr2[1]=binrule[arr1[j-1]][arr1[0]][arr1[1]][arr1[2]][arr1[3]];
  for (i=2; i<j-2; i++) 
    arr2[i]=binrule[arr1[i-2]][arr1[i-1]][arr1[i]][arr1[i+1]][arr1[i+2]];
  arr2[j-2]=binrule[arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]];
  arr2[j-1]=binrule[arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]];
} else if (j==4) {
  arr2[0]=binrule[arr1[2]][arr1[3]][arr1[0]][arr1[1]][arr1[2]]; 
  arr2[1]=binrule[arr1[3]][arr1[0]][arr1[1]][arr1[2]][arr1[3]]; 
  arr2[2]=binrule[arr1[0]][arr1[1]][arr1[2]][arr1[3]][arr1[0]]; 
  arr2[3]=binrule[arr1[1]][arr1[2]][arr1[3]][arr1[0]][arr1[1]]; 
} else if (j==3) {
  arr2[0]=binrule[arr1[1]][arr1[2]][arr1[0]][arr1[1]][arr1[2]]; 
  arr2[1]=binrule[arr1[2]][arr1[0]][arr1[1]][arr1[2]][arr1[0]]; 
  arr2[2]=binrule[arr1[0]][arr1[1]][arr1[2]][arr1[0]][arr1[1]]; 
} else if (j==2) {
  arr2[0]=binrule[arr1[0]][arr1[1]][arr1[0]][arr1[1]][arr1[0]]; 
  arr2[1]=binrule[arr1[1]][arr1[0]][arr1[1]][arr1[0]][arr1[1]]; 
} else if (j==1) {
  arr2[0]=binrule[arr1[0]][arr1[0]][arr1[0]][arr1[0]][arr1[0]]; 
}
if (co==1) for (ii=0; ii<j; ii++) arr1[(int)ii]=arr2[(int)(ii)];
if (co==0) for (ii=0; ii<j; ii++) arr1[(int)ii]=arr2[(int)((7*ii)%j)];
if (co==2) for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];}
if (co==3) for (ii=0; ii<j; ii+=3) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];
				 arr1[(int)ii+2]=arr2[(int)((7*ii+2)%j)];}
if (co==4) for (ii=0; ii<j; ii+=4) {arr1[(int)ii]=arr2[(int)((7*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((7*ii+1)%j)];
				 arr1[(int)ii+2]=arr2[(int)((7*ii+2)%j)];
				 arr1[(int)ii+3]=arr2[(int)((7*ii+3)%j)];}

if (co==5) {for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((11*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((11*ii+1)%j)];}
	    for (ii=0; ii<j; ii+=8) {
		k=arr1[(int)ii];
		arr1[(int)ii]=arr1[(int)((ii+2)%j)];
		arr1[(int)((ii+2)%j)]=k;
		k=arr1[(int)((ii+1)%j)];
		arr1[(int)((ii+1)%j)]=arr1[(int)((ii+4)%j)];
		arr1[(int)((ii+4)%j)]=k;
		k=arr1[(int)((ii+3)%j)];
		arr1[(int)((ii+3)%j)]=arr1[(int)((ii+5)%j)];
		arr1[(int)((ii+5)%j)]=k;
                }
	    }
if (co==6) {for (ii=0; ii<j; ii+=2) {arr1[(int)ii]=arr2[(int)((11*ii)%j)];
				 arr1[(int)ii+1]=arr2[(int)((11*ii+1)%j)];}
	    for (ii=0; ii<j; ii+=16) {
		k=arr1[(int)ii];
		arr1[(int)ii]=arr1[(int)((ii+2)%j)];
		arr1[(int)((ii+2)%j)]=k;
		k=arr1[(int)((ii+1)%j)];
		arr1[(int)((ii+1)%j)]=arr1[(int)((ii+4)%j)];
		arr1[(int)((ii+4)%j)]=k;
		k=arr1[(int)((ii+3)%j)];
		arr1[(int)((ii+3)%j)]=arr1[(int)((ii+5)%j)];
		arr1[(int)((ii+5)%j)]=k;

		k=arr1[(int)((ii+8)%j)];
		arr1[(int)((ii+8)%j)]=arr1[(int)((ii+10)%j)];
		arr1[(int)((ii+10)%j)]=k;
		k=arr1[(int)((ii+9)%j)];
		arr1[(int)((ii+9)%j)]=arr1[(int)((ii+15)%j)];
		arr1[(int)((ii+15)%j)]=k;
		k=arr1[(int)((ii+11)%j)];
		arr1[(int)((ii+11)%j)]=arr1[(int)((ii+13)%j)];
		arr1[(int)((ii+13)%j)]=k;
		k=arr1[(int)((ii+12)%j)];
		arr1[(int)((ii+12)%j)]=arr1[(int)((ii+14)%j)];
		arr1[(int)((ii+14)%j)]=k;
		}
	    } break;

 /* a radius 1 von Neumann neighborhood */

case 2: if (j<16384) break;
	  for (i0=0; i0<128; i0++) for (i1=0; i1<128; i1++) 
	    arr2[ix2(i0,i1)]=binrule
	      [arr1[ix2(i0,i1+1)]]
	      [arr1[ix2(i0-1,i1)]]
	      [arr1[ix2(i0,i1)]]
	      [arr1[ix2(i0+1,i1)]]
	      [arr1[ix2(i0,i1-1)]];
	  for (i=0; i<16384; i++) arr1[i]=arr2[i];
	    break;

 /* a tetrahedron with a cell in the center */

case 3: if (j<16384) break;
	  for (i0=0; i0<32; i0++)
	  for (i1=0; i1<32; i1++)
	  for (i2=0; i2<16; i2++)
	    arr2[ix3(i0,i1,i2)]=binrule
	      [arr1[ix3(i0,i1+1,i2+1)]]
 	      [arr1[ix3(i0,i1-1,i2+1)]]
	      [arr1[ix3(i0,i1,i2)]]
	      [arr1[ix3(i0+1,i1,i2-1)]]
	      [arr1[ix3(i0-1,i1,i2-1)]];
	  for (i=0; i<16384; i++) arr1[i]=arr2[i];
	    break;

/* four one-sided neighbors and their corner cell */

case 4: if (j<16384) break;
	  for (i0=0; i0<16; i0++) for (i1=0; i1<8; i1++) 
	  for (i2=0; i2<16; i2++) for (i3=0; i3<8; i3++) 
	    arr2[ix4(i0,i1,i2,i3)]=binrule
	      [arr1[ix4(i0+1,i1,i2,i3)]]
	      [arr1[ix4(i0,i1+1,i2,i3)]]
	      [arr1[ix4(i0,i1,i2,i3)]]
	      [arr1[ix4(i0,i1,i2+1,i3)]]
	      [arr1[ix4(i0,i1,i2,i3+1)]];
	  for (i=0; i<16384; i++) arr1[i]=arr2[i];
	    break;

 /* five one-sided neighbors without the corner cell */

case 5: if (j<16384) break;
	  for (i0=0; i0<8; i0++) for (i1=0; i1<8; i1++) 
	  for (i2=0; i2<8; i2++) for (i3=0; i3<8; i3++) 
	  for (i4=0; i4<4; i4++)
	    arr2[ix5(i0,i1,i2,i3,i4)]=binrule
	      [arr1[ix5(i0+1,i1,i2,i3,i4)]]
	      [arr1[ix5(i0,i1+1,i2,i3,i4)]]
	      [arr1[ix5(i0,i1,i2+1,i3,i4)]]
	      [arr1[ix5(i0,i1,i2,i3+1,i4)]]
	      [arr1[ix5(i0,i1,i2,i3,i4+1)]];
	  for (i=0; i<16384; i++) arr1[i]=arr2[i];
	    break;

default: break;}
}


/* evolution for one generation 
onegen(j) int j; {int i;
  if (j<2) return;
  arr2[0]=binrule[arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]][arr1[2]];
  arr2[1]=binrule[arr1[j-1]][arr1[0]][arr1[1]][arr1[2]][arr1[3]];
 
 for (i=2; i<j-2; i++)
  arr2[ i ]=binrule[arr1[i-2]][arr1[i-1]][arr1[i]][arr1[i+1]][arr1[i+2]];
  arr2[j-2]=binrule[arr1[j-4]][arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]];
  arr2[j-1]=binrule[arr1[j-3]][arr1[j-2]][arr1[j-1]][arr1[0]][arr1[1]];


  for (i=0; i<j; i++) arr1[i]=arr2[i];
}
 */


/* index two dimensions */
int ix2(i,j) int i, j; {return (128+i)%128+128*((128+j)%128);}

/* index three dimensions */
int ix3(i,j,k) int i, j, k; {return (32+i)%32+ 32*((32+j)%32+ 32*((16+k)%16 ));}

/* index four dimensions */
int ix4(i,j,k,l) int i, j, k, l; {
return (16+i)%16+ 16*( (8+j)%8+ 8*( (16+k)%16+ 16*( (8+l)%8 )));}

/* index five dimensions */
int ix5(i,j,k,l,m) int i, j, k, l, m; {
return (8+i)%8+ 8*( (8+j)%8+ 8*( (8+k)%8+ 8*( (8+l)%8+ 8*( (4+m)%4 ))));}

/* generate line of cells in arr1 with n mils of 1's */
rprob(n) int n; {int i; for (i=0; i<AL; i++) arr1[i]=(random()%1000<n)?1:0;}

/* generate line of n cells in character array a with p mils of 1's */
ranchay(a,p,n) char *a; int p, n; {int i;
  for (i=0; i<n; i++) a[i]=(random()%1000<p)?'1':'0';}

/* generate a random line of cells in arr1 */
ranlin() {int i; for (i=0; i<AL; i++) arr1[i]=random()%KK;}

/* generate a random core for a planar array */
ranplacor() {int i, j;
  for (i=60; i<69; i++) for (j=60; j<69; j++)  arr1[ix2(i,j)] = random()%KK;}

/* generate a random core for a solid array */
ransolcor() {int i, j, k;
  for (i=14; i<19; i++) for (j=14; j<19; j++) for (k=7; k<10; k++) 
	 arr1[ix3(i,j,k)] = random()%KK;}

/* clear auxrule to blanks */
auxblnk(x) char x[DS]; {int i; for (i=0; i<DS; i++) x[i]=' ';}

/* clear ascrule to zeroes */
asczero(x) char x[DS]; {int i; for (i=0; i<DS; i++) x[i]='0';}

/* increment ascrule except where protected by auxrule */
ascincr(x,y) char x[DS], y[DS]; {int i;
for (i=0; i<DS; i++) {
  if (y[i]!=' ') continue;
  if (x[i]=='0') {x[i]='1'; break;}
  if (x[i]=='1') x[i]='0';
  }}

/* uniform line menu */
ulmenu(l) int l; {int i, j, k;
  for (j=0; j*SL<YL; j++) {
  for (i=0; i<SL;    i++) {
    k=arr1[j*SL+i]+'0';
    videocursor(0,LROW+j,i); videocattr(0,k=='0'?'.':k,l,1);
    }}
  }

/* tutorial and Help screen */
tuto() {
    videomode(T80X25);
    videocursor(0,2,0);
    scrstr("<Copyright (C) 1987, 1988, 1989 - H.V. McIntosh, G. Cisneros S.>");
    videocursor(0,4,0);
    scrstr("           ***** LIFE in One Dimension *****");
    videocursor(0,6,0);
    scrstr("Two States - Black(0), Cyan(1).");
    videocursor(0,8,0);
    scrstr("Second neighbors - two on each side, five altogether.");
    videocursor(0,10,0);
    scrstr("Complete transition rule - random, edited, or stored.");
    videocursor(0,12,0);
    scrstr("Initial Cellular Array - random, edited, or patterned.");
    videocursor(0,14,0);
    scrstr("Submenus in options t(probabilities) and d(de Bruijn)");
    videocursor(0,15,0);
    scrstr("will be displayed in response to typing ?");
    videocursor(0,17,0);
    scrstr("Use any key to terminate a display in progress.");
    videocursor(0,21,0);
    scrstr("Now, ... press any key to continue.");
}

/* rule menu */
rmenu() {
    videocursor(0,CR-5,0);
    scrstr("      00000000000000001111111111111111");
    videocursor(0,CR-4,0);
    scrstr("      00000000111111110000000011111111");
    videocursor(0,CR-3,0);
    scrstr("      00001111000011110000111100001111");
    videocursor(0,CR-2,0);
    scrstr("      00110011001100110011001100110011");
    videocursor(0,CR-1,0);
    scrstr("      01010101010101010101010101010101");
    videocursor(0,CR,0);
    scrstr("Rule: ");
    scrrul();
    videocursor(0,CR,42);   printf("         ");
    videocursor(0,CR+1,42); printf("         ");
    videocursor(0,CR,42);
    if (istot()==1) printf("(T-%2d)",totonu(0));
    videocursor(0,CR+1,42);
    if (isstot()==1) printf("(S:%2d,%2d)",stotonu(0,0),stotonu(0,1));
    videocursor(0,CM,0);
    scrstr("    r(rule), l(line), #nn(stored), g(graph), q(quit), t(prob),");
    videocursor(0,CM+1,0);
    scrstr("         x(rand rule), y(rand line), u(unit line), d(deBruijn)");
    videocursor(0,CM+2,0);
    scrstr("         @nn(tot/rule), $nn(12 tot/rules), T(ed tot/rule).");
    videocursor(0,CM,0);
    }

/* totalistic rule menu*/
tmenu() {
    videocursor(0,CR-3,50);
    scrstr("      0....1");
    videocursor(0,CR-2,50);
    scrstr("rule: ");
    tscrrul();
    videocursor(0,CR-2,56);
    }

/* semitotalistic rule menu*/
smenu() {
    videocursor(0,CR-3,50); scrstr("       01234");
    videocursor(0,CR-2,50); scrstr("born:  "); sscrrul(0);
    videocursor(0,CR-1,50); scrstr("live:  "); sscrrul(1);
    videocursor(0,CR-2,57);
    }

/* line menu */
lmenu() {int i, j, k;
  for (j=0; 40*j<YL; j++) {
    videocursor(0,CL+j,0);
    for (i=0; i<40; i++) {
      k=arr1[40*j+i];
      videoputc(k==0?'.':k+'0',1);
      }
    }
  videocursor(0,CM,0);
  }

/* display Wolfram rule number */
wmenu(n) int n;
{char xx[4]; int i, nn;  
    nn=sprintf(xx,"%3d",n);
    videocursor(0,2,40);
    for (i=0; i<nn; i++) videoputc(xx[i],1);
    videocursor(0,2,40); }

/* display rule number */
xmenu(n) int n;
{char xx[4]; int i, nn;  
    nn=sprintf(xx,"%3d",n);
    videocursor(0,1,40);
    for (i=0; i<nn; i++) videoputc(xx[i],1);
    videocursor(0,1,40); }

/* clear screen space for rule numbers */
xblnk() {
    videocursor(0,CR-4,40);
    scrstr("     ");
    videocursor(0,CR-3,40);
    scrstr("     ");
    videocursor(0,CR,0); }

/* copy saved rule #n into active rule */
xtoasc(asc,n) char asc[KK][KK][KK][KK][KK]; int n; {
 char a[DS]; int i, i0, i1, i2, i3, i4;
//xmenu(n+1);
for (i=0; i<DS; i++) a[i]=ixrule[n][i];
          i=0;
            for (i0=0; i0<KK; i0++){
	      for (i1=0; i1<KK; i1++){
		for (i2=0; i2<KK; i2++){ 
  		   for(i3=0; i3<KK; i3++) {
		      for(i4=0; i4<KK; i4++){
			 asc[i0][i1][i2][i3][i4]=a[i];
			    i++;
					     }
					   }
					}
				     }
			          }


}

/* change semitotalistic rule to general rule */
sttoasc() {int i0, i1, i2, i3, i4;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
ascrule[i0][i1][i2][i3][i4]=strule[i2][i0+i1+i3+i4];
}

/* change totalistic rule to general rule */
totoasc() {int i0, i1, i2, i3, i4;
for (i0=0; i0<KK; i0++){
    for (i1=0; i1<KK; i1++){
       for (i2=0; i2<KK; i2++){
           for (i3=0; i3<KK; i3++){
             for (i4=0; i4<KK; i4++){
                  ascrule[i0][i1][i2][i3][i4]=ttrule[i0+i1+i2+i3+i4];
				     }
				  }
     			      }
        		   }
  			}
}

/* change decimal totalistic rule to sum values */
nutoto(x) int x; {int i; for (i=0; i<TS; i++) {ttrule[i]=x%KK+'0'; x/=KK;}}

/* change sum values to decimal semitotalistic rule */
int stotonu(i,j) int i, j; {int r;
 r=0; if (i<TS-1) r=(strule[j][i]-'0')+KK*stotonu(i+1,j);
  return r; }

/* change sum values to decimal totalistic rule */
int totonu(i) int i; {int r;
 r=0; if (i<TS) r=(ttrule[i]-'0')+KK*totonu(i+1);
  return r; }

/* test whether a rule is semi-totalistic */
int isstot() {int st[KK][TS-1], i0, i1, i2, i3, i4, l;
    l=1;
    st[0][0]=ascrule[0][0][0][0][0];
    st[0][1]=ascrule[1][0][0][0][0];
    st[0][2]=ascrule[1][1][0][0][0];
    st[0][3]=ascrule[1][1][0][1][0];
    st[0][4]=ascrule[1][1][0][1][1];
    st[1][0]=ascrule[0][0][1][0][0];
    st[1][1]=ascrule[1][0][1][0][0];
    st[1][2]=ascrule[1][1][1][0][0];
    st[1][3]=ascrule[1][1][1][1][0];
    st[1][4]=ascrule[1][1][1][1][1];
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
    if (ascrule[i0][i1][i2][i3][i4]!=st[i2][i0+i1+i3+i4]) l=0;
    return l; }

/* test whether a rule is totalistic */
int istot() {int tr[TS], i0, i1, i2, i3, i4, l;
    l=1;
    tr[0]=ascrule[0][0][0][0][0];
    tr[1]=ascrule[1][0][0][0][0];
    tr[2]=ascrule[1][1][0][0][0];
    tr[3]=ascrule[1][1][1][0][0];
    tr[4]=ascrule[1][1][1][1][0];
    tr[5]=ascrule[1][1][1][1][1];
    for (i0=0; i0<KK; i0++)
    for (i1=0; i1<KK; i1++)
    for (i2=0; i2<KK; i2++)
    for (i3=0; i3<KK; i3++)
    for (i4=0; i4<KK; i4++)
    if (ascrule[i0][i1][i2][i3][i4]!=tr[i0+i1+i2+i3+i4]) l=0;      
    return l; }

/* find the totalistic rule number of a known ascii totalistic rule */
void asctotot() {int i0, i1, i2, i3, i4;
for (i0=0; i0<KK; i0++)
for (i1=0; i1<KK; i1++)
for (i2=0; i2<KK; i2++)
for (i3=0; i3<KK; i3++)
for (i4=0; i4<KK; i4++)
      ttrule[i0+i1+i2+i3+i4] = ascrule[i0][i1][i2][i3][i4];
}

/* put a single i-cell on a line of j-cells */
void putionj(i,j) int i, j; 
{int k; for (k=0; k<AL; k++) arr1[k]=j; arr1[AL/2]=i;}

/* make all of arr1 equal to zero */
zerline() {int i; for (i=0; i<AL; i++) arr1[i]=0;}

/* limit j to interval (i,k) */
int lim(i,j,k) int i, j, k;
    {if (i>=j) return i; if (k<=j) return k; return j;}

/* display the rule number on the screen */
scrrul() {int i0, i1, i2, i3, i4;
  for (i0=0; i0<KK; i0++)
  for (i1=0; i1<KK; i1++)
  for (i2=0; i2<KK; i2++)
  for (i3=0; i3<KK; i3++)
  for (i4=0; i4<KK; i4++)
  videoputc(ascrule[i0][i1][i2][i3][i4],1);      
  }

/* display semitotalistic rule number by sum */
sscrrul(j) int j; {int i; for (i=0; i<TS-1; i++) videoputc(strule[j][i],1); }

/* display the rule number on the screen in glorious technicolor */
cscrrul() {int i, i0, i1, i2, i3, i4;
  i=6;
  for (i0=0; i0<KK; i0++)
  for (i1=0; i1<KK; i1++)
  for (i2=0; i2<KK; i2++)
  for (i3=0; i3<KK; i3++)
  for (i4=0; i4<KK; i4++) {
  videocursor(0,UROW,i++);
  videocattr(0,ascrule[i0][i1][i2][i3][i4],auxrule[i0][i1][i2][i3][i4]==' '?3:6,1);
  } }

/* turn ascrule into its Wolfram number and place on screen at (i,j) */
woruno(i,j) int i, j; {
char xx[12];
int  k, l, i0, i1, i2, i3, i4;
  k=0;
  for (i0=KK-1; i0>=0; i0--)
  for (i1=KK-1; i1>=0; i1--)
  for (i2=KK-1; i2>=0; i2--)
  for (i3=KK-1; i3>=0; i3--)
  for (i4=KK-1; i4>=0; i4--)
    k=KK*k+(ascrule[i0][i1][i2][i3][i4]-'0');
  k=sprintf(xx,"%10d",k);
  videocursor(0,i,j);
  for (l=0; l<k; l++) videoputc(xx[l],1);
}

/* report Wolfram rule number or variant */
int wolnum(c) char c; {
int  k, i0, i1, i2, i3, i4;
  k=0;
  switch (c) {
    case 'w':
      for (i4=KK-1; i4>=0; i4--)
      for (i3=KK-1; i3>=0; i3--)
      for (i2=KK-1; i2>=0; i2--)
      for (i1=KK-1; i1>=0; i1--)
      for (i0=KK-1; i0>=0; i0--)
        k=2*k+(ascrule[i0][i1][i2][i3][i4]-'0');
      break;
    case 'c':
      for (i4=KK-1; i4>=0; i4--)
      for (i3=KK-1; i3>=0; i3--)
      for (i2=KK-1; i2>=0; i2--)
      for (i1=KK-1; i1>=0; i1--)
      for (i0=KK-1; i0>=0; i0--)
        k=2*k+((ascrule[(i0+1)%KK][(i1+1)%KK][(i2+1)%KK][(i3+1)%KK][(i4+1)%KK]-'0')+1)%KK;
      break;
    case 'r':
      for (i4=KK-1; i4>=0; i4--)
      for (i3=KK-1; i3>=0; i3--)
      for (i2=KK-1; i2>=0; i2--)
      for (i1=KK-1; i1>=0; i1--)
      for (i0=KK-1; i0>=0; i0--)
        k=2*k+(ascrule[i4][i3][i2][i1][i0]-'0');
      break;
    case 'b':
      for (i4=KK-1; i4>=0; i4--)
      for (i3=KK-1; i3>=0; i3--)
      for (i2=KK-1; i2>=0; i2--)
      for (i1=KK-1; i1>=0; i1--)
      for (i0=KK-1; i0>=0; i0--)
        k=2*k+((ascrule[(i4+1)%KK][(i3+1)%KK][(i2+1)%KK][(i1+1)%KK][(i0+1)%KK]-'0')+1)%KK;
      break;
    default: k=0; break;}
        return k;
}

txtmode() {videomode(T80X25); rmenu(); lmenu();}
grfmode() {videomode(COLGRAF);}

/* display totalistic rule number by sum */
tscrrul() {int i; for (i=0; i<TS; i++) {videoputc(ttrule[i],1); } }

/* change decimal wolfram rule to binary values */
nutowo(x,pr) char pr[KK*KK*KK]; int x; {int i;
for (i=0; i<KK*KK*KK; i++) {pr[i]=x%KK+'0'; x/=KK;}}

/* write a string in graphics mode */
scrstr(s) char *s; {for (; *s != '\0'; s++) videoputc(*s,1); } 

/* place time and date on screen */ 
tidat() {long cl[1]; char *ctime(); time(cl); printf("%26s",ctime(cl));}

/* keyboard status */
int kbdst() {return(bdos(11) & 0xFF);}

/* direct keyboard input, no echo */
kbdin() {int c;
  if ((c = bdos(7) & 0xFF) == '\0') c = (bdos(7) & 0xFF) | 0x80;
  return(c);
}

/* turn a nibble into an ASCII character */
int nibasc(i) int i; {if (i<10) return '0'+i; else return 'A'+i-10;}

/* turn ASCII character into hexadecimal value */
int nibbl(c) int c; {
  if (c>='0'&&c<='9') return (c-'0');
  if (c>='a'&&c<='f') return (c-'a'+10);
  if (c>='A'&&c<='F') return (c-'A'+10);
  return 0;
}

/* read number from keyboard */
int numin(n) int n; {char c;
  c=kbdin();
  if (c>='0'&&c<='9') return(numin(10*n+(c-'0'))); else return(n);
}

void rule22(){
    xtoasc(ascrule,(rand()%NX));
    ranlin();						/* random initial array */
}

void savearry() {int i; for (i=0; i<AL; i++) arr3[i]=arr1[i];}
void renewarry() {int i; for (i=0; i<AL; i++) arr1[i]=arr3[i];}

int  bdos(){}

/* end nxlcau22.c */
